#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <malloc.h>
#include <io.h>
#include <dos.h>
#include <direct.h>
#include <signal.h>
#include <process.h>
#include "intspy.h"



/* The name of the file loggings have to be written to. */
char *log_file_name;

/* A table of interrupts to be monitored, and the section name of the
description. */
struct SPY_TAB *spy_tab;

/* An array of tables (one for each interrupt handler), in which is stored 
what to print under what condition. */
struct INT_TAB *int_tab[MAX_INT_HANDLERS];

/* Default value for spy_tab. (The interrupt number is filled in later.) */
struct SPY_TAB default_spy_tab[] = {
	{ -1, DEFAULT_DESCRIPTION_SECTION } ,
	{ -1, NULL }
};

/* A flag indicating we are executing our own code. If we are inside our
own code we will always chain to the orriginal interrupt handler. */
int in_our_own_code=1;

/* A pointer to the orriginal interrupt handler. */
void (_interrupt _far *old_int_handler[MAX_INT_HANDLERS])();

/* The stack pointer we will load in our own interrupt handler. */
char _huge *int_handler_stack_pointer;

/* The position where the original stack pointer is saved, when we set
up our own stack. */
char _huge *old_stack_pointer;


/* Variables to which the register values are copied from the stack,
before we execute our own interrupt functions. If our own interrupt
functions are ready, they are copied back to the stack. */
unsigned int res;
unsigned int rds;
unsigned int rdi;
unsigned int rsi;
unsigned int rbp;
unsigned int rsp;
unsigned int rbx;
unsigned int rdx;
unsigned int rcx;
unsigned int rax;
unsigned int rip;
unsigned int rcs;
unsigned int rflags;
unsigned int rss;

/* The number of the interrut we are processing. */
unsigned int current_interrupt_no;



int main(int argc, char *argv[])
{
        /* An array which will be used as stack for our own interrupt
        handler. */
        char int_handler_stack[INT_HANDLER_STACK_SIZE];

        /* The value returned by the spawnvp function. */
        int return_value;

        /* A file pointer used for the initinal opening of the log
        file. */
        FILE *log_fp;

	/* The name of the ini file section in which is specified which
	interrupts we are going to log. */
	char ini_file_section[STRING_SIZE];

	/* The ordinal number of the command line parameter specifying the
	command to be executed. */
	int command_arg_no;

        /* A counter. */
        int i;

	/* If first and only argument is "-?", "/?" or "--help",
	print copyright notice and help information. */
	if(argc==2 && (argv[1][0]=='-'||argv[1][0]=='/') && 
		argv[1][1]=='?' && argv[1][2]=='\0' || argc==2 && 
		argv[1][0]=='-' && argv[1][1]=='-' && 
		argv[1][2]=='h' && argv[1][3]=='e' && 
		argv[1][4]=='l' && argv[1][5]=='p' && 
		argv[1][6]=='\0')
        {
#ifdef LIMMITED_VERSION
                fprintf(stdout,"%s %s (limmited version)   %s\n\n%s",
                        CMDNAME,VERSION,COPYRIGHT,HELP);
#else
                fprintf(stdout,"%s %s   %s\n\n%s",
                        CMDNAME,VERSION,COPYRIGHT,HELP);
#endif
                exit(0);
        }

	/* First parameter is -setup, second parameter is logdev, and
	third parameter is command. */
	if( (argv[1][0]=='-' || argv[1][0]=='/') && argv[2][0]!='-' && 
		argv[2][0]!='/' && 4<=argc )
	{
		strcpy(ini_file_section,SETUP_SECTION_PREFIX);
		strncat(ini_file_section,argv[1]+1,STRING_SIZE-
			strlen(SETUP_SECTION_PREFIX)-1);
		sscanf(argv[1]+1,"%x",&default_spy_tab[0].interrupt_no);
		command_arg_no=3;
		log_file_name=argv[2];
	}

	/* First parameter is logdev, second parameter is -setup, and
	third parameter is command. */
	else if( (argv[2][0]=='-' || argv[2][0]=='/') && argv[1][0]!='-' && 
		argv[1][0]!='/' && 4<=argc )
	{
		strcpy(ini_file_section,SETUP_SECTION_PREFIX);
		strncat(ini_file_section,argv[2]+1,STRING_SIZE-
			strlen(SETUP_SECTION_PREFIX)-1);
		sscanf(argv[2]+1,"%x",&default_spy_tab[0].interrupt_no);
		command_arg_no=3;
		log_file_name=argv[1];
	}

	/* First parameter is logdev, and second parameter is command. */
	else if(argv[1][0]!='-' && argv[1][0]!='/' && 
		argv[2][0]!='-' && argv[2][0]!='/' && 3<=argc)
	{
		strcpy(ini_file_section,SETUP_SECTION_PREFIX);
		strncat(ini_file_section,DEFAULT_SETUP_SECTION_SUFFIX,
			STRING_SIZE-strlen(SETUP_SECTION_PREFIX)-1);
		command_arg_no=2;
		log_file_name=argv[1];
	}

	/* Otherwise print an error message and die. */
	else 
        {
                fprintf(stderr,USAGE);
                exit(1);
        }

        /* Read the interupt function desriptions from the ini file. */
        if(read_spy_tab(INI_FILE_NAME,ini_file_section,&spy_tab)<=0)
	{
		if(default_spy_tab[0].interrupt_no<0 || 
			default_spy_tab[0].interrupt_no>0xff)
		{
			fprintf(stderr,"%s: Section [%s] is empty or file %s not found\n",
				CMDNAME,ini_file_section,INI_FILE_NAME);
			exit(1);
		}
		spy_tab=default_spy_tab;
	}
        for(i=0;i<MAX_INT_HANDLERS && spy_tab[i].interrupt_no>=0;i++)
        {
                if(read_int_tab(INI_FILE_NAME,
			spy_tab[i].section,&int_tab[i])<=0)
		{
			fprintf(stderr,"%s: Section [%s] is empty or file %s not found\n",
				CMDNAME,spy_tab[i].section,INI_FILE_NAME);
			exit(1);
		}
        }

	/* The side effect of this call is that any open ini file is closed. */
	get_private_profile_string("","","",NULL,0,"");

	/* This first call to dos_busy() initializes the indos flag pointer
	and the crittical error flag pointer. */
	dos_busy();

#ifdef LIMMITED_VERSION

	/* If this version is a limmited version, only interrupt 21h
	may be monitored. */
        for(i=0;i<MAX_INT_HANDLERS && spy_tab[i].interrupt_no>=0;i++)if(
		spy_tab[i].interrupt_no!=0x21)
	{
		fprintf(stderr,LIMMITED_VERSION);
		exit(1);
	}

#endif

        /* Open the log device. */
        if( (log_fp=fopen(log_file_name,"wt"))==NULL )
        {
                fprintf(stderr,"%s: Can't open %s\n",
                        CMDNAME,log_file_name);
                exit(1);
        }

        /* Close it again. */
        fclose(log_fp);

        /* Initialise the stack pointer which will be used by our own
        interrupt handler. The stack grows downward, so the stack
        pointer must initianally point to the last element of int_handler_stack.
        Because int_handler_stack is a dynamic variable, it resides on the
        current stack. If the new stack would be somewhere else, the
        stack checking routine used by many standard library functions,
        thinks we have run out of stack space. */
        int_handler_stack_pointer=int_handler_stack;
        int_handler_stack_pointer+=INT_HANDLER_STACK_SIZE-1;

        /* To be sure we are able restore the original interrupt 
        vector before exit, we disable control-c, and other signals. */
        signal(SIGABRT,SIG_IGN);
        signal(SIGINT,SIG_IGN);
        signal(SIGTERM,SIG_IGN);

	/* Install our own interrupt handlers. */
        for(i=0;i<MAX_INT_HANDLERS && spy_tab[i].interrupt_no>=0;i++)
        {
                /* Save old interrupt vector. */
                old_int_handler[i]=_dos_getvect(spy_tab[i].interrupt_no);

                /* Set interrupt to our own handler. */
                switch(i)
                {
                    case 0:
                        _dos_setvect(spy_tab[i].interrupt_no,new_int_handler_0);
                        break;
                    case 1:
                        _dos_setvect(spy_tab[i].interrupt_no,new_int_handler_1);
                        break;
                    case 2:
                        _dos_setvect(spy_tab[i].interrupt_no,new_int_handler_2);
                        break;
                    case 3:
                        _dos_setvect(spy_tab[i].interrupt_no,new_int_handler_3);
                        break;
                    case 4:
                        _dos_setvect(spy_tab[i].interrupt_no,new_int_handler_4);
                        break;
                    case 5:
                        _dos_setvect(spy_tab[i].interrupt_no,new_int_handler_5);
                        break;
                    case 6:
                        _dos_setvect(spy_tab[i].interrupt_no,new_int_handler_6);
                        break;
                    case 7:
                        _dos_setvect(spy_tab[i].interrupt_no,new_int_handler_7);
                        break;
                }
        }

	/* Indicate we are no longer executing our own code. */
        in_our_own_code=0;

        /* Run the command specified at the command line as a child process.
        If this command exits the routine spawnvp returns. */
        return_value=spawnvp(P_WAIT, argv[command_arg_no], 
		&argv[command_arg_no]);

	/* Indicate we are nagain executing our own code. */
        in_our_own_code=1;

	/* Restore orriginal interrupt vectors. */
        for(i--;i>=0;i--)
        {
                /* Restore original interrupt vector. */
                _dos_setvect(spy_tab[i].interrupt_no,old_int_handler[i]);
        }

	/* The side effect of this statement is that any buffered loggings
	are flushed. */
	lprintf("");

        /* If spawn failed, restore the original interrupt vector,
        give an error message, and die. */
        if(return_value==-1)
        {
                /* Error message. */
                fprintf(stderr,
                        "%s: Unable to spawn %s\n",
			CMDNAME,argv[command_arg_no]);

                /* Return to DOS with return value -1. */
                exit(1);
        }

        /* Return to DOS with return value. */
        return(return_value);

}




