#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <malloc.h>
#include <io.h>
#include <dos.h>
#include <direct.h>
#include <signal.h>
#include <process.h>
#include "intspy.h"




/* Do all loggings. Int_tab describes what to log under which condition. */
void spy(struct INT_TAB int_tab[])
{
	int entry_no;

	/* Scan all enties in int_tab. */
	for(entry_no=0;int_tab[entry_no].description[0]!=NULL;entry_no++)
	{
		/* If the conditions are met, fall through. */
		if(test_condition(int_tab[entry_no].condition)==0)break;
	}

	/* Log to file. */
	if(int_tab[entry_no].description[0]!=NULL)lprintf(
		get_first_par(int_tab[entry_no].description[0]),
		get_par(int_tab[entry_no].description[1]),
		get_par(int_tab[entry_no].description[2]),
		get_par(int_tab[entry_no].description[3]),
		get_par(int_tab[entry_no].description[4]),
		get_par(int_tab[entry_no].description[5]),
		get_par(int_tab[entry_no].description[6]),
		get_par(int_tab[entry_no].description[7]));
	else lprintf("");

	/* Ready. */
	return;
}





/* Return 0 if the register contents match the values in the array
condition. Otherwise return 1. */
int test_condition(int condition[])
{
	if( condition[0]!=-1 && condition[0]!=(rax>>8) )return 1;
	if( condition[1]!=-1 && condition[1]!=(rax&0xff) )return 1;
	if( condition[2]!=-1 && condition[2]!=(rbx>>8) )return 1;
	if( condition[3]!=-1 && condition[3]!=(rbx&0xff) )return 1;
	if( condition[4]!=-1 && condition[4]!=(rcx>>8) )return 1;
	if( condition[5]!=-1 && condition[5]!=(rcx&0xff) )return 1;
	if( condition[6]!=-1 && condition[6]!=(rdx>>8) )return 1;
	if( condition[7]!=-1 && condition[7]!=(rdx&0xff) )return 1;
	return 0;
}




/* Get a pointer to the first parameter. Leading and trailing '"' are
stripped. */
char *get_first_par(char *parameter)
{
	static char str[STRING_SIZE];
	int i,j;

	/* If the parameter is the NULL pointer, return it. */
	if(parameter==NULL)return NULL;

	/* Copy parameter without leading '"'. */
	for(i=1,j=0;j<127 && parameter[i]!='\0';i++,j++)
	{
		str[j]=parameter[i];
	}
	str[j]='\0';

	/* If parameter terminates with a '"', strip it. */
	if(str[j-1]=='"')str[j-1]='\0';

	/* Return the copied parameter. */
	return str;

}



/* Get the value of a parameter. An unsigned integer can be returned, but 
also a near pointer casted into an unsigned integer can be returned. */
unsigned int get_par(char *parameter)
{
	/* A circulair buffer of storage for strings. */
	static char string_table[16][STRING_SIZE];

	/* Index into string_table, indicating the next free slot. */
	static unsigned int string_number=0;

	/* A pointer created out of a pair of register values. */
	char _huge *string_pointer;

	/* Counters. */
	int i,j;

	/* String_number must be between 0 and 15. */
	string_number&=0xf;

	/* If the parameter is the NULL pointer, return 0. */
	if(parameter==NULL)return 0;

	/* If the parameter describes a register, return the contents of
	that register. */
	if(case_independant_strcmp(parameter,"ax")==0)return rax;
	if(case_independant_strcmp(parameter,"ah")==0)return rax>>8;
	if(case_independant_strcmp(parameter,"al")==0)return rax&0xff;
	if(case_independant_strcmp(parameter,"bx")==0)return rbx;
	if(case_independant_strcmp(parameter,"bh")==0)return rbx>>8;
	if(case_independant_strcmp(parameter,"bl")==0)return rbx&0xff;
	if(case_independant_strcmp(parameter,"cx")==0)return rcx;
	if(case_independant_strcmp(parameter,"ch")==0)return rcx>>8;
	if(case_independant_strcmp(parameter,"cl")==0)return rcx&0xff;
	if(case_independant_strcmp(parameter,"dx")==0)return rdx;
	if(case_independant_strcmp(parameter,"dh")==0)return rdx>>8;
	if(case_independant_strcmp(parameter,"dl")==0)return rdx&0xff;
	if(case_independant_strcmp(parameter,"ds")==0)return rds;
	if(case_independant_strcmp(parameter,"es")==0)return res;
	if(case_independant_strcmp(parameter,"cs")==0)return rcs;
	if(case_independant_strcmp(parameter,"ss")==0)return rss;
	if(case_independant_strcmp(parameter,"di")==0)return rdi;
	if(case_independant_strcmp(parameter,"si")==0)return rsi;
	if(case_independant_strcmp(parameter,"bp")==0)return rbp;
	if(case_independant_strcmp(parameter,"sp")==0)return rsp;
	if(case_independant_strcmp(parameter,"ip")==0)return rip;
	if(case_independant_strcmp(parameter,"flags")==0)return rflags;

	/* If the parameter is intno, return the interrupt number. */
	if(case_independant_strcmp(parameter,"intno")==0)
		return current_interrupt_no;

	/* If the parameter is "a_" followed by the description of a pair
	of registers, create a huge pointer out of the two register values. */
	if(parameter[0]=='a' && parameter[1]=='_')
	{
		if(case_independant_strcmp(parameter,"a_dsax")==0)
			string_pointer=make_huge_pointer(rds,rax);
		else if(case_independant_strcmp(parameter,"a_dsbx")==0)
			string_pointer=make_huge_pointer(rds,rbx);
		else if(case_independant_strcmp(parameter,"a_dscx")==0)
			string_pointer=make_huge_pointer(rds,rcx);
		else if(case_independant_strcmp(parameter,"a_dsdx")==0)
			string_pointer=make_huge_pointer(rds,rdx);
		else if(case_independant_strcmp(parameter,"a_dsdi")==0)
			string_pointer=make_huge_pointer(rds,rdi);
		else if(case_independant_strcmp(parameter,"a_dssi")==0)
			string_pointer=make_huge_pointer(rds,rsi);
		else if(case_independant_strcmp(parameter,"a_esax")==0)
			string_pointer=make_huge_pointer(res,rax);
		else if(case_independant_strcmp(parameter,"a_esbx")==0)
			string_pointer=make_huge_pointer(res,rbx);
		else if(case_independant_strcmp(parameter,"a_escx")==0)
			string_pointer=make_huge_pointer(res,rcx);
		else if(case_independant_strcmp(parameter,"a_esdx")==0)
			string_pointer=make_huge_pointer(res,rdx);
		else if(case_independant_strcmp(parameter,"a_esdi")==0)
			string_pointer=make_huge_pointer(res,rdi);
		else if(case_independant_strcmp(parameter,"a_essi")==0)
			string_pointer=make_huge_pointer(res,rsi);
		else if(case_independant_strcmp(parameter,"a_csax")==0)
			string_pointer=make_huge_pointer(rcs,rax);
		else if(case_independant_strcmp(parameter,"a_csbx")==0)
			string_pointer=make_huge_pointer(rcs,rbx);
		else if(case_independant_strcmp(parameter,"a_cscx")==0)
			string_pointer=make_huge_pointer(rcs,rcx);
		else if(case_independant_strcmp(parameter,"a_csdx")==0)
			string_pointer=make_huge_pointer(rcs,rdx);
		else if(case_independant_strcmp(parameter,"a_csdi")==0)
			string_pointer=make_huge_pointer(rcs,rdi);
		else if(case_independant_strcmp(parameter,"a_cssi")==0)
			string_pointer=make_huge_pointer(rcs,rsi);
		else if(case_independant_strcmp(parameter,"a_ssax")==0)
			string_pointer=make_huge_pointer(rss,rax);
		else if(case_independant_strcmp(parameter,"a_ssbx")==0)
			string_pointer=make_huge_pointer(rss,rbx);
		else if(case_independant_strcmp(parameter,"a_sscx")==0)
			string_pointer=make_huge_pointer(rss,rcx);
		else if(case_independant_strcmp(parameter,"a_ssdx")==0)
			string_pointer=make_huge_pointer(rss,rdx);
		else if(case_independant_strcmp(parameter,"a_ssdi")==0)
			string_pointer=make_huge_pointer(rss,rdi);
		else if(case_independant_strcmp(parameter,"a_sssi")==0)
			string_pointer=make_huge_pointer(rss,rsi);
	}

	/* If the parameter is "h_" followed by the description of a pair
	of registers, create a huge pointer out of the two register values. */
	else if(parameter[0]=='h' && parameter[1]=='_')
	{
		if(case_independant_strcmp(parameter,"h_dsax")==0)
			string_pointer=make_huge_pointer(rds,rax);
		else if(case_independant_strcmp(parameter,"h_dsbx")==0)
			string_pointer=make_huge_pointer(rds,rbx);
		else if(case_independant_strcmp(parameter,"h_dscx")==0)
			string_pointer=make_huge_pointer(rds,rcx);
		else if(case_independant_strcmp(parameter,"h_dsdx")==0)
			string_pointer=make_huge_pointer(rds,rdx);
		else if(case_independant_strcmp(parameter,"h_dsdi")==0)
			string_pointer=make_huge_pointer(rds,rdi);
		else if(case_independant_strcmp(parameter,"h_dssi")==0)
			string_pointer=make_huge_pointer(rds,rsi);
		else if(case_independant_strcmp(parameter,"h_esax")==0)
			string_pointer=make_huge_pointer(res,rax);
		else if(case_independant_strcmp(parameter,"h_esbx")==0)
			string_pointer=make_huge_pointer(res,rbx);
		else if(case_independant_strcmp(parameter,"h_escx")==0)
			string_pointer=make_huge_pointer(res,rcx);
		else if(case_independant_strcmp(parameter,"h_esdx")==0)
			string_pointer=make_huge_pointer(res,rdx);
		else if(case_independant_strcmp(parameter,"h_esdi")==0)
			string_pointer=make_huge_pointer(res,rdi);
		else if(case_independant_strcmp(parameter,"h_essi")==0)
			string_pointer=make_huge_pointer(res,rsi);
		else if(case_independant_strcmp(parameter,"h_csax")==0)
			string_pointer=make_huge_pointer(rcs,rax);
		else if(case_independant_strcmp(parameter,"h_csbx")==0)
			string_pointer=make_huge_pointer(rcs,rbx);
		else if(case_independant_strcmp(parameter,"h_cscx")==0)
			string_pointer=make_huge_pointer(rcs,rcx);
		else if(case_independant_strcmp(parameter,"h_csdx")==0)
			string_pointer=make_huge_pointer(rcs,rdx);
		else if(case_independant_strcmp(parameter,"h_csdi")==0)
			string_pointer=make_huge_pointer(rcs,rdi);
		else if(case_independant_strcmp(parameter,"h_cssi")==0)
			string_pointer=make_huge_pointer(rcs,rsi);
		else if(case_independant_strcmp(parameter,"h_ssax")==0)
			string_pointer=make_huge_pointer(rss,rax);
		else if(case_independant_strcmp(parameter,"h_ssbx")==0)
			string_pointer=make_huge_pointer(rss,rbx);
		else if(case_independant_strcmp(parameter,"h_sscx")==0)
			string_pointer=make_huge_pointer(rss,rcx);
		else if(case_independant_strcmp(parameter,"h_ssdx")==0)
			string_pointer=make_huge_pointer(rss,rdx);
		else if(case_independant_strcmp(parameter,"h_ssdi")==0)
			string_pointer=make_huge_pointer(rss,rdi);
		else if(case_independant_strcmp(parameter,"h_sssi")==0)
			string_pointer=make_huge_pointer(rss,rsi);
	}

	/* If the parameter describes a pair of registers, create a huge 
	pointer out of the two register values. */
	else if(case_independant_strcmp(parameter,"dsax")==0)
		string_pointer=make_huge_pointer(rds,rax);
	else if(case_independant_strcmp(parameter,"dsbx")==0)
		string_pointer=make_huge_pointer(rds,rbx);
	else if(case_independant_strcmp(parameter,"dscx")==0)
		string_pointer=make_huge_pointer(rds,rcx);
	else if(case_independant_strcmp(parameter,"dsdx")==0)
		string_pointer=make_huge_pointer(rds,rdx);
	else if(case_independant_strcmp(parameter,"dsdi")==0)
		string_pointer=make_huge_pointer(rds,rdi);
	else if(case_independant_strcmp(parameter,"dssi")==0)
		string_pointer=make_huge_pointer(rds,rsi);
	else if(case_independant_strcmp(parameter,"esax")==0)
		string_pointer=make_huge_pointer(res,rax);
	else if(case_independant_strcmp(parameter,"esbx")==0)
		string_pointer=make_huge_pointer(res,rbx);
	else if(case_independant_strcmp(parameter,"escx")==0)
		string_pointer=make_huge_pointer(res,rcx);
	else if(case_independant_strcmp(parameter,"esdx")==0)
		string_pointer=make_huge_pointer(res,rdx);
	else if(case_independant_strcmp(parameter,"esdi")==0)
		string_pointer=make_huge_pointer(res,rdi);
	else if(case_independant_strcmp(parameter,"essi")==0)
		string_pointer=make_huge_pointer(res,rsi);
	else if(case_independant_strcmp(parameter,"csax")==0)
		string_pointer=make_huge_pointer(rcs,rax);
	else if(case_independant_strcmp(parameter,"csbx")==0)
		string_pointer=make_huge_pointer(rcs,rbx);
	else if(case_independant_strcmp(parameter,"cscx")==0)
		string_pointer=make_huge_pointer(rcs,rcx);
	else if(case_independant_strcmp(parameter,"csdx")==0)
		string_pointer=make_huge_pointer(rcs,rdx);
	else if(case_independant_strcmp(parameter,"csdi")==0)
		string_pointer=make_huge_pointer(rcs,rdi);
	else if(case_independant_strcmp(parameter,"cssi")==0)
		string_pointer=make_huge_pointer(rcs,rsi);
	else if(case_independant_strcmp(parameter,"ssax")==0)
		string_pointer=make_huge_pointer(rss,rax);
	else if(case_independant_strcmp(parameter,"ssbx")==0)
		string_pointer=make_huge_pointer(rss,rbx);
	else if(case_independant_strcmp(parameter,"sscx")==0)
		string_pointer=make_huge_pointer(rss,rcx);
	else if(case_independant_strcmp(parameter,"ssdx")==0)
		string_pointer=make_huge_pointer(rss,rdx);
	else if(case_independant_strcmp(parameter,"ssdi")==0)
		string_pointer=make_huge_pointer(rss,rdi);
	else if(case_independant_strcmp(parameter,"sssi")==0)
		string_pointer=make_huge_pointer(rss,rsi);


	/* If the parameter is a string, return a copy of this string 
	without the leading and trailing '"'. */
	else if(parameter[0]=='"')
	{
		/* Copy parameter without leading '"'. */
		for(i=1,j=0;j<STRING_SIZE-1 && parameter[i]!='\0';i++,j++)
			string_table[string_number][j]=parameter[i];
		string_table[string_number][j]='\0';

		/* Strip terminating '"'. */
		if(string_table[string_number][j-1]=='"')
			string_table[string_number][j-1]='\0';

		/* Return pointer to string, and increase the string 
		number. */
		return (unsigned int)string_table[string_number++];
	}

	/* In all other cases return 0. */
	else return 0;

	/* If we are at this point, the parameter describes a pair of
	registers, and string_pointer is a huge pointer made out of
	the two register values. */

	/* If the description of the pair of registers is prefixed
	with "h_", the string pointed to by string_pointer must be
	converted to hex. */
	if(parameter[0]=='h' && parameter[1]=='_')
	{
		for(i=0;i<(STRING_SIZE-1)/3 && i<H_STRING_SIZE-1;i++)
		{
			sprintf(string_table[string_number]+3*i,
				"%02.2x ",(unsigned char)(string_pointer[i]));
		}
		string_table[string_number][3*i-1]='\0';
	}

	/* If the description of the pair of registers is prefixed
	with "a_", the string pointed to by string_pointer must be
	copied. Any non printing character must be 
	converted into '*'. */
	else if(parameter[0]=='a' && parameter[1]=='_')
	{
		for(i=0;i<STRING_SIZE-1 && i<A_STRING_SIZE-1;i++)
		{
			if(0<=string_pointer[i] && string_pointer[i]<32)
				string_table[string_number][i]='*';
			else string_table[string_number][i]=string_pointer[i];
		}
		string_table[string_number][i]='\0';
	}

	/* Otherwise the string pointed to by string_pointer must be copied
	up to the first zero byte. Any non printing character must be 
	converted into '*'. */
	else
	{
		for(i=0;i<STRING_SIZE-1 && string_pointer[i]!='\0';i++)
		{
			if(0<=string_pointer[i] && string_pointer[i]<32)
				string_table[string_number][i]='*';
			else string_table[string_number][i]=string_pointer[i];
		}
		string_table[string_number][i]='\0';
	}

	/* Return pointer to string. This pointer is casted two an unsigned 
	int. Because we use the small memory model, we can convert this
	value back into a pointer later. Increase string_number, so
	it indicates the next free slot. */
	return (unsigned int)string_table[string_number++];
}



/* Construct a _huge pointer from a segment address and an offset. */
void _huge *make_huge_pointer(unsigned int segment,unsigned int offset) 
{
	unsigned long int bit_pattern;

	bit_pattern=segment;
	bit_pattern<<=16;
	bit_pattern|=offset;

	return (void _huge *)bit_pattern;
}





/* Opens the file named log_file_name, print to it, and close it again. 
The syntax of lprintf() is equal to the syntax of printf(). 
The dots int the function definition mean a variable argument list.
For a description of this feature, see Kernighan and Ritchie, The
C Programming Language, ANSI C edition. */
void lprintf(char *fmt, ...)
{
	static char buffer[PRINT_BUFFER_SIZE]="";
	static int buffer_overflow=0;
	va_list args;
	FILE *log_fp;

	/* If there is nothing to write to the log file, just return. */
	if(fmt[0]=='\0' && buffer[0]=='\0')return;

	/* If DOS is busy, or it is not possible to open the
	log file, we write our logging into a buffer, so
	it can be written to the file later. */
	if(dos_busy() || (log_fp=fopen(log_file_name,"a"))==NULL )
	{
		/* IF the buffer already overflowed, just return. */
		if(buffer_overflow)return;

		/* If there is not enough space left in our buffer, put
		an overflow message in the buffer, set the overflow
		flag, and return. */
		if(strlen(buffer)+1>=PRINT_BUFFER_SIZE-MAX_SPY_MESSAGE_LENGTH)
		{
			/* Overflow message. */
			sprintf(buffer,
				"\n\n%s: Buffer overflow, spy messages lost\n\n",
				CMDNAME);

			/* Set overflow flag. */
			buffer_overflow=1;

			/* Ready. */
			return;
		}

		/* Append logging to buffer. */
		va_start(args,fmt);
		vsprintf(buffer+strlen(buffer),fmt,args);
		va_end(args);

		/* Ready. */
		return;
	}

	/* If there is something in the buffer, write it to the file. */
	if(buffer[0]!='\0')
	{
		/* Write to file. */
		fprintf(log_fp,"%s",buffer);

		/* Empty buffer. */
		buffer[0]='\0';

		/* Reset overflow flag. */
		buffer_overflow=0;
	}

	/* Write logging message to log file. */
	va_start(args,fmt);
	vfprintf(log_fp,fmt,args);
	va_end(args);

	/* Close log file. */
	fclose(log_fp);
}


/* Return non-zero if DOS is busy. The first time this funtion is called,
it initializes the indos flag pointer and the crittical error flag pointer.
This must be done at a moment we are sure dos is not busy. */
int dos_busy(void)
{
	static char huge *indos_flag_pointer=NULL;
	static char huge *critical_error_flag_pointer=NULL;

	union REGS regs;
	struct SREGS sregs;

	/* If we haven't yet got a pointer to the indos flag and
	the crittical error flag, get them. */
	if(indos_flag_pointer==NULL)
	{
		/* Get pointer to indos flag. */
		regs.h.ah=0x34;
		intdosx(&regs,&regs,&sregs);
		indos_flag_pointer=make_huge_pointer(sregs.es,regs.x.bx);

		/* Get pointer to crittical error flag. */
		regs.x.ax=0x5d06;
		intdosx(&regs,&regs,&sregs);
		critical_error_flag_pointer=make_huge_pointer(sregs.ds,
			regs.x.si);
	}

	/* If indos flag or crittical error flag are set, return 1, 
	otherwise return 0. */
	return *indos_flag_pointer || *critical_error_flag_pointer;
}


