#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include "intspy.h"



/* Allocate memory for spy_tab, and fill it wit the items found in
ini_file_section in ini_file_name. Return the number of items found. */
int read_spy_tab(char ini_file_name[],char ini_file_section[],
	struct SPY_TAB *(*spy_tab))
{

	int i,j;

	char entry_list[1024];
	char *entry_list_pointer;
	int entry_count;
	int entry_no;

	char description_line[256];
	char *description_line_pointer;

	/* Get a list of items in the requested section. */
	get_private_profile_string(ini_file_section,NULL,"",entry_list,1024,
		ini_file_name);

	/* Count the entries in entry_list. */
	entry_count=0;
	if(entry_list[0]!=0)
	{
		for(i=0;entry_list[i]!='\0' || entry_list[i+1]!='\0';i++)
			if(entry_list[i]=='\0')entry_count++;
		entry_count++;
	}

	/* If no entries found return 0. */
	if(entry_count==0)return 0;

	/* If there are to many entries, exit with an error message. */
	if(entry_count>MAX_INT_HANDLERS)
	{
		fprintf(stderr,
			"%s: To many entries found in section [%s] in %s\n",
			CMDNAME,ini_file_section,ini_file_name);
		exit(1);
	}

	/* Allocate memory for spy_tab. */
	*spy_tab=malloc(sizeof(struct SPY_TAB)*(entry_count+1));
	if(*spy_tab==NULL)
	{
		fprintf(stderr,"%s: Out of memory\n",CMDNAME);
		exit(1);
	}

	/* Initialize the allocated memory. */
	for(entry_no=0;entry_no<entry_count+1;entry_no++)
	{	
		(*spy_tab)[entry_no].interrupt_no=-1;
		(*spy_tab)[entry_no].section=NULL;
	}

	/* Let entry_list_pointer point to the first entry. */
	entry_list_pointer=entry_list;

	/* Process all entiries. */
	for(entry_no=0;entry_no<entry_count && *entry_list_pointer!='\0';
		entry_no++)
	{

		/* The entry must represent the interrupt number. 
		If the entry is not hexadecimal, exit with an error 
		message. */
		if(sscanf(entry_list_pointer,"%x",
			&(*spy_tab)[entry_no].interrupt_no)!=1)
		{
			fprintf(stderr,
				"%s: Bad entry %s found in section [%s] in %s\n",
				CMDNAME,entry_list_pointer,
				ini_file_section,ini_file_name);
			exit(1);
		}

		/* Get the parameters for the entry. */
		get_private_profile_string(ini_file_section,entry_list_pointer,
			DEFAULT_DESCRIPTION_SECTION,
			description_line,256,ini_file_name);

		/* Strip leading space. */
		for(i=0;description_line[i]==' ' || description_line[i]=='\t';
			i++);

		/* Find first ',' or '\0'. */
		for(j=i;description_line[j]!=',' && description_line[j]!='\0';
			j++);

		/* Strip everything after the ','. */
		description_line[j]='\0';

		/* The description section is the first parameter. If no 
		description section name is specified, set it to the default 
		value. */
		if(j==i)
		{
			(*spy_tab)[entry_no].section=
				DEFAULT_DESCRIPTION_SECTION;
		}

		/* Else allocate memory and copy found section name to it. */
		else
		{

			/* Allocate memory for the description section name. */
			(*spy_tab)[entry_no].section=
				malloc(strlen(description_line+i)+1);
			if((*spy_tab)[entry_no].section==NULL)
			{
				fprintf(stderr,"%s: Out of memory\n",CMDNAME);
				exit(1);
			}

			/* Copy description section name to the allocated 
			memory. */
			for(j=0;description_line[i]!=',' && 
				description_line[i]!='\0';i++,j++)
				(*spy_tab)[entry_no].section[j]=
				description_line[i];
			(*spy_tab)[entry_no].section[j]='\0';
		}


		/* Find the next entry. */
		while(*entry_list_pointer!=0)entry_list_pointer++;
		entry_list_pointer++;
	}

	/* Return the number of entries. */
	return entry_count;
}

