#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include "intspy.h"



/* Allocate memory for int_tab, and fill it wit the items found in
ini_file_section in ini_file_name. Return the number of items found. */
int read_int_tab(char ini_file_name[],char ini_file_section[],
	struct INT_TAB *(*int_tab))
{

	int i,j,k;

	char entry_list[1024];
	char *entry_list_pointer;
	int entry_count;
	int entry_no;
	int par_no;
	int condition_no;
	int in_string;

	char description_line[256];
	char *description_line_pointer;

	/* Get a list of items in the requested section. */
	get_private_profile_string(ini_file_section,NULL,"",entry_list,1024,
		ini_file_name);

	/* Count the entries in entry_list. */
	entry_count=0;
	if(entry_list[0]!=0)
	{
		for(i=0;entry_list[i]!='\0' || entry_list[i+1]!='\0';i++)
			if(entry_list[i]=='\0')entry_count++;
		entry_count++;
	}

	/* If no entries found return 0. */
	if(entry_count==0)return 0;

	/* Allocate memory for spy_tab. */
	*int_tab=malloc(sizeof(struct INT_TAB)*(entry_count+1));
	if(*int_tab==NULL)
	{
		fprintf(stderr,"%s: Out of memory\n",CMDNAME);
		exit(1);
	}

	/* Initialize the allocated memory. */
	for(entry_no=0;entry_no<entry_count+1;entry_no++)
	{	
		for(par_no=0;par_no<8;par_no++)
			(*int_tab)[entry_no].description[par_no]=NULL;
		for(condition_no=0;condition_no<8;condition_no++)
			(*int_tab)[entry_no].condition[condition_no]=-1;
	}

	/* Let entry_list_pointer point to the first entry. */
	entry_list_pointer=entry_list;

	/* Process all entries. */
	for(entry_no=0;entry_no<entry_count && *entry_list_pointer!='\0';
		entry_no++)
	{

		/* The entry name acts as a pattern which must match the
		register values. The values are copied to the condition
		fields. A -1 represents a wildcard. Since all condition
		fields are initialized to -1, all fields not explicitly
		filled in are wildcards. */
		for(condition_no=0;condition_no<8;condition_no++)
		{
			/* End of string reached. The rest of the condition
			fields represent wildcards. */
			if(entry_list_pointer[2*condition_no]=='\0')break;

			/* If there are question marks in the entry name,
			these represent wildcards. */
			if(entry_list_pointer[2*condition_no]=='?' &&
				entry_list_pointer[2*condition_no+1]=='?')
				continue;

			/* If there are non-hexadecimal characters in the
			entry (except for the wilrdcards), print an error 
			message and die. */
			if( char_to_hex(entry_list_pointer[2*condition_no])<0 ||
				char_to_hex(
				entry_list_pointer[2*condition_no+1])<0 )
			{
				fprintf(stderr,
					"%s: Bad entry %s in section [%s] in %s\n",
					CMDNAME,entry_list_pointer,
					ini_file_section,ini_file_name);
				exit(1);
			}

			/* Convert the hexadecimal digits in the entry name,
			into an integer value and store it into the 
			corresponding condition field. */
			(*int_tab)[entry_no].condition[condition_no]=
				char_to_hex(entry_list_pointer[2*condition_no]);
			(*int_tab)[entry_no].condition[condition_no]*=16;
			(*int_tab)[entry_no].condition[condition_no]+=
				char_to_hex(entry_list_pointer[2*condition_no+1]);

		}

		/* Get the parameters for the entry. */
		get_private_profile_string(ini_file_section,entry_list_pointer,
			"",description_line,256,ini_file_name);

		/* Let description_line_pointer point to the first character
		of the description. */
		description_line_pointer=description_line;

		/* Allocate memory for the description, and assign it 
		to int_tab...... */
		(*int_tab)[entry_no].description[0]=
			malloc(strlen(description_line)+1);
		if((*int_tab)[entry_no].description[0]==NULL)
		{
			fprintf(stderr,"%s: Out of memory\n",CMDNAME);
			exit(1);
		}

		/* At the beginning we are not inside a string. */
		in_string=0;

		/* The first parameter is numbered 0 */
		par_no=0;

		/* I is the index into the description. We start at the 
		beginning. */
		i=0;

		/* Process all parameters. */
		for(;;)
		{
			/* If a '"' found, toggle the value of in_string. */
			if(*description_line_pointer=='"')in_string^=1;	

			/* End of parameter found. */
			if(*description_line_pointer=='\0')
			{
				(*int_tab)[entry_no].
					description[par_no][i]='\0';

				/* If the first parameter does not start
				with a '"', print an error message and die. */
				if((*int_tab)[entry_no].description[0][0]!='"')
				{
					fprintf(stderr,
						"%s: Bad first parameter for entry %s in section [%s] in file %s\n",
						CMDNAME,entry_list_pointer,
						ini_file_section,ini_file_name);
					exit(1);
				}

				/* Strip trailing space. */
				for(;i>0 && ((*int_tab)[entry_no].description
					[par_no][i-1]==' ' ||
					(*int_tab)[entry_no].description
					[par_no][i-1]=='\t');i--)
					(*int_tab)[entry_no].description
					[par_no][i-1]='\0';

				/* If a parameter begins with a quote, it
				must also end with a quote. */
				if(i-1>=0 && (*int_tab)[entry_no].description
					[par_no][i-1]!='"' &&
					(*int_tab)[entry_no].description
					[par_no][0]=='"')
				{
					fprintf(stderr,
						"%s: Bad parameter for entry %s in section [%s] in %s\n",
						CMDNAME,entry_list_pointer,
						ini_file_section,ini_file_name);
					exit(1);
				}

				/* Ready with this item. */
				break;
			}

			/* A comma outside a string, terminates the
			parameter. */
			else if(*description_line_pointer==',' && !in_string)
			{
				/* Replace the ',' by a '\0'. */
				(*int_tab)[entry_no].description[par_no][i]='\0';
				/* If the first parameter does not start
				with a '"', print an error message and die. */
				if((*int_tab)[entry_no].description[0][0]!='"')
				{
					fprintf(stderr,
						"%s: Bad first parameter for entry %s in section [%s] in %s\n",
						CMDNAME,entry_list_pointer,
						ini_file_section,ini_file_name);
					exit(1);
				}

				/* Strip trailing space. */
				for(;i>0 && ((*int_tab)[entry_no].description
					[par_no][i-1]==' ' ||
					(*int_tab)[entry_no].description
					[par_no][i-1]=='\t');i--)
					(*int_tab)[entry_no].description
					[par_no][i-1]='\0';
				
				/* If a parameter begins with a quote but
				must also end with a quote. */
				if(i-1>=0 && (*int_tab)[entry_no].description
					[par_no][i-1]!='"' &&
					(*int_tab)[entry_no].description
					[par_no][0]=='"')
				{
					fprintf(stderr,
						"%s: Bad parameter for entry %s in section [%s] in %s\n",
						CMDNAME,entry_list_pointer,
						ini_file_section,ini_file_name);
					exit(1);
				}

				/* If we have read 8 parameters, break. */
				if(7<=par_no)break;

				/* Increment the parameter number. */
				par_no++;

				/* Assign a pointer to the next parameter
				to int_tab.... */
				(*int_tab)[entry_no].description[par_no]=
					&((*int_tab)[entry_no].
					description[par_no-1][i+1]);

				/* Set the index to 0 again. */
				i=0;

				/* Let description_line_pointer point to 
				the next parameter. */
				description_line_pointer++;

				/* Continue with the next character. */
				continue;
			}

			/* If we find a '\\' inside a sting it is, together
			with the folowing character, transfered into something
			else. */
			else if(*description_line_pointer=='\\' && in_string)
			{
				/* Let description_line_pointer point to the
				character following the '\\'. */
				description_line_pointer++;

				/* What is written to int_tab.... depends on
				the next character. */
				switch(*description_line_pointer)
				{

				    /* "\n" is a new-line. */
				    case 'n':
				    	(*int_tab)[entry_no].
						description[par_no][i]='\n';
				    	i++;
				        description_line_pointer++;
					break;

				    /* "\t" is a tab. */
				    case 't':
				    	(*int_tab)[entry_no].
						description[par_no][i]='\t';
				    	i++;
				        description_line_pointer++;
					break;

				    /* "\v" is a vertical tab. */
				    case 'v':
				    	(*int_tab)[entry_no].
						description[par_no][i]='\v';
				    	i++;
				        description_line_pointer++;
					break;

				    /* "\b" is a backspace. */
				    case 'b':
				    	(*int_tab)[entry_no].
						description[par_no][i]='\b';
				    	i++;
				        description_line_pointer++;
					break;

				    /* "\r" is a cariage-return. */
				    case 'r':
				    	(*int_tab)[entry_no].
						description[par_no][i]='\r';
				    	i++;
				        description_line_pointer++;
					break;

				    /* "\f" is a formfeed. */
				    case 'f':
				    	(*int_tab)[entry_no].
						description[par_no][i]='\f';
				    	i++;
				        description_line_pointer++;
					break;

				    /* "\a" is a audible bell. */
				    case 'a':
				    	(*int_tab)[entry_no].
						description[par_no][i]='\a';
				    	i++;
				        description_line_pointer++;
					break;

				    /* "\x.." is a hexadecimal character
				    value. */
				    case 'x':

					/* First digit. */
				        description_line_pointer++;
					if(char_to_hex(
						*description_line_pointer)<0)
						break;
					(*int_tab)[entry_no].
						description[par_no][i]=
						char_to_hex(
						*description_line_pointer);
					description_line_pointer++;

					/* Second digit. */
					if(char_to_hex(
						*description_line_pointer)<0)
					{
						i++;
						break;
					}
					(*int_tab)[entry_no].
						description[par_no][i]*=16;
					(*int_tab)[entry_no].
						description[par_no][i]+=
						char_to_hex(
						*description_line_pointer);
					description_line_pointer++;

					/* Ready. */
				    	i++;
					break;

				    /* "\..." is a octal character value. */
				    case '0':
				    case '1':
				    case '2':
				    case '3':
				    case '4':
				    case '5':
				    case '6':
				    case '7':

				    	/* First digit. */
					(*int_tab)[entry_no].
						description[par_no][i]=
						*description_line_pointer-'0';
					description_line_pointer++;

				    	/* Second digit. */
					if(*description_line_pointer<'0' ||
						'7'< *description_line_pointer)
					{
						i++;
						break;
					}
					(*int_tab)[entry_no].
						description[par_no][i]*=8;
					(*int_tab)[entry_no].
						description[par_no][i]+=
						*description_line_pointer-'0';
					description_line_pointer++;

					/* Third digit. */
					if(*description_line_pointer<'0' ||
						'7'< *description_line_pointer)
					{
						i++;
						break;
					}
					(*int_tab)[entry_no].
						description[par_no][i]*=8;
					(*int_tab)[entry_no].
						description[par_no][i]+=
						*description_line_pointer-'0';
					description_line_pointer++;

					/* Ready. */
				    	i++;
					break;

				    /* "\\" is a backspace. */
				    case '\\':
				    	(*int_tab)[entry_no].
						description[par_no][i]='\\';
				    	i++;
				        description_line_pointer++;
					break;

				    /* "\"" is a quote. */
				    case '"':
				    	(*int_tab)[entry_no].
						description[par_no][i]='"';
				    	i++;
				        description_line_pointer++;
					break;
				}

				/* Continue with the next character. */
				continue;
			}
			else 
			{
				/* Add any characters, except leading
				space to int_tab..... */
				if(i!=0 || (*description_line_pointer!=' '
					&& *description_line_pointer!='\t'))
				{
					(*int_tab)[entry_no].
						description[par_no][i]= 
						*description_line_pointer;
					i++;
				}

				/* Let description_line_pointer point to 
				the next parameter. */
				description_line_pointer++;

				/* Continue with the next character. */
				continue;
			}

		}

		/* Find the next entry. */
		while(*entry_list_pointer!=0)entry_list_pointer++;
		entry_list_pointer++;

	}

	/* Return the number of entries. */
	return entry_count;
}



/* Convert a character into a hexadecimal number between 0 and 15. On error
return -1. */
int char_to_hex(char c)
{
	if('0'<=c && c<='9')return c-'0';
	if('a'<=c && c<='f')return c-'a'+10;
	if('A'<=c && c<='F')return c-'a'+10;
	return -1;
}
