#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "intspy.h"





/* Get parameters for item item in section section from ini file ini_file_name.
Result is a pointer to the memory location at which the result must be stored.
Size_of_result is the maximum number of characters (including the terminating
'\0') to be copied to result. Default_result is the default value. If item
is NULL, a list of items in section is copied to result. Return the actual
size of the result. */
int get_private_profile_string(char section[],char item[],char default_result[],
	char *result,int size_of_result,char ini_file_name[])
{

	static char last_ini_file_name[STRING_SIZE];
        static FILE *ini_file_pointer=NULL;        
        static char section_name[STRING_SIZE]="";

	int ini_file_was_open;

        char item_name[STRING_SIZE]="";
        char line[STRING_SIZE];
        char *server_pars;
        char dummy[STRING_SIZE];
	char path[PATH_SIZE];
	char *path_table[PATH_TABLE_SIZE];
	char ini_file_path_name[STRING_SIZE];
	int path_table_counter;
        int i;
        int j;
        int c;
	int current_size_of_result;

	/* If the ini filename is the null string, just close any open 
	file. */
	if(ini_file_name[0]=='\0')
	{
		/* If there is an ini file open, close it. */
		if(ini_file_pointer!=NULL)
		{
			fclose(ini_file_pointer);
			ini_file_pointer=NULL;
		}

		/* Mark the ini file was not yet open. */
		ini_file_was_open=0;
	}

	/* If no ini file is open, or if the open ini file is the wrong file, 
	open the ini file. */
	else if( ini_file_pointer==NULL || case_independant_strcmp(
		last_ini_file_name,ini_file_name)!=0 )
	{
		/* If there is another ini file open, close it. */
		if(ini_file_pointer!=NULL)
		{
			fclose(ini_file_pointer);
			ini_file_pointer=NULL;
		}

		/* Remember the ini filename. */
		strncpy(last_ini_file_name,ini_file_name,STRING_SIZE-1);
		last_ini_file_name[STRING_SIZE-1]='\0';

		/* Mark the ini file was not yet open. */
		ini_file_was_open=0;

		/* Fill path table with the current directory and all 
		directories from PATH. */
		if(getenv(PATH_ENVIRONMENT_VAR)!=NULL)
		{
			/* Get the path. */
			strncpy(path,getenv(PATH_ENVIRONMENT_VAR),PATH_SIZE-1);
			path[PATH_SIZE-1]='\0';

			/* The first directory to be searched is the current 
			directory. */
			path_table[0]=CURRENT_DIRECTORY;

			/* Fill the rest of path_table with the directories from 
			the environment variable PATH. */
			for(path_table_counter=1,i=0;
				path_table_counter<(PATH_TABLE_SIZE-1);
				path_table_counter++,i++)
			{
			
				/* Skip delimmiters in PATH. */ 
				while(path[i]==PATH_ENVIRONMENT_DELIMITER ||
					path[i]==PATH_ENVIRONMENT_DELIMITER_2 ||
					path[i]==PATH_ENVIRONMENT_DELIMITER_3)
					i++;

				/* If we reached the end of the PATH, break. */
				if(path[i]=='\0')break;

				/* Assign a pointer the directory found to 
				path_table. */
				path_table[path_table_counter]=path+i;

				/* Find the next delimmiter. */
				while(path[i]!=PATH_ENVIRONMENT_DELIMITER && 
					path[i]!=PATH_ENVIRONMENT_DELIMITER_2 &&
					path[i]!=PATH_ENVIRONMENT_DELIMITER_3 &&
					path[i]!='\0')i++;

				/* Replace the delimmiter by a '\0'. */
				path[i]='\0';
			}

			/* Terminate path_table with a NULL pointer. */
			path_table[path_table_counter]=NULL;
		}

		/* If the environment variable path does not exist, fill path
		table only with the current directory. */
		else
		{
			path_table[0]=CURRENT_DIRECTORY;
			path_table[1]=NULL;
		}

		/* Try to open the ini file for all directories in 
		path_table. */
		for(path_table_counter=0;path_table[path_table_counter]!=NULL;
			path_table_counter++)
		{
			/* Copy directory name to ini_file_path_name. */
			strncpy(ini_file_path_name,
				path_table[path_table_counter],STRING_SIZE-1);
			ini_file_path_name[STRING_SIZE-1]='\0';

			/* Find terminating '\0'. */
			for(i=0;ini_file_path_name[i]!='\0';i++);

			/* If the last characteris not a directory name 
			separator, add one. */
			if(i>0 && ini_file_path_name[i-1]!=
				DIRECTORY_NAME_SEPARATOR && 
				ini_file_path_name[i-1]!=
				DIRECTORY_NAME_SEPARATOR_2 && 
				ini_file_path_name[i-1]!=
				DIRECTORY_NAME_SEPARATOR_3)
			{
				ini_file_path_name[i]=DIRECTORY_NAME_SEPARATOR;
				if(i<STRING_SIZE-1)i++;
				ini_file_path_name[i]='\0';
			}

			/* Finally add the name of the ini file. */
			strncat(ini_file_path_name,ini_file_name,
				STRING_SIZE-strlen(ini_file_path_name)-1);

			/* Try to open the ini file. */
			ini_file_pointer=fopen(ini_file_path_name,"r");

			/* On succes break. */
			if(ini_file_pointer!=NULL)break;
		}

		/* If the ini file is not opened, return the default result. */
		if( ini_file_pointer==NULL )
		{
			if(item==NULL)return current_size_of_result;
			else return strlen(result);
		}

		/* At the beginning of the file, we haven't found any section
		name yet. */
		section_name[0]='\0';
	}

	/* Otherwise we will continue reading from the previously opened ini
	file. Set a flag to indicate this. */
	else
	{
		ini_file_was_open=1;
	}

	/* If a list of item in section is requested. */
	if(item==NULL)
	{
		/* Nothing has been found yet. */
		current_size_of_result=0;
		result[0]='\0';

		/* Start at the befinning of the file. */
		rewind(ini_file_pointer);

		/* Clear the flag we are reading from a previously opened
		file, because we just rewinded the file. */
		ini_file_was_open=0;

		/* At the beginning of the file, we haven't found any section
		name yet. */
		section_name[0]='\0';
	}

	/* If the parameters for an item are requested. */
	else
	{
		/* Copy default_result to result. If an item is found in the 
		ini file, the result is overwritten. */
		strncpy(result,default_result,size_of_result);
		if(size_of_result>0)result[size_of_result-1]='\0';
	}

	/* If opening ini file failed, return default. */
	if(ini_file_pointer==NULL)
	{
		/* If a list of items in a section was requested, there is 
		no default, so return 0 as size of the result. */
		if(item==NULL)return 0;

		/* Otherwise result contains the specified default value. */
		else return strlen(result);
	}

        /* Looping until end-of-file. */
        for(;;)
        {
                /* Eat white space. */
                for(;;)
                {
                        c=getc(ini_file_pointer);
                        if(c!=' ' && c!='\t' && c!='\n' && c!='\r' &&
				c!='\f' && c!='\v')
                        {
                                if(c!=EOF)ungetc(c,ini_file_pointer);
                                break;
                        }
                }

		/* On end-of-file. */
                if(c==EOF)
                {
			/* If we are reading from a previously opened
			ini file, start again from the beginning of 
			the file. */
			if(ini_file_was_open)
			{
				/* Start again from the beginning of the 
				file. */
				rewind(ini_file_pointer);

				/* At the beginning of the file we are not
				inside a section. */
				section_name[0]='\0';

				/* Clear the mark we are reading from a 
				previously opened file. The next time we
				reach end-of-file, we won't rewind again. */
				ini_file_was_open=0;

				/* Continue reading from the rewinded file. */
				continue;
			}

			/* Otherwise we have already read the whole file. */
			else
			{
				/* If a list of items in a section was 
				requested, result contains the items found
				so far. Size of result contains the size 
				of the result. */
				if(item==NULL)return current_size_of_result;

				/* Otherwise result contains the 
				specified default value. */
				else return strlen(result);
			}
                }

                /* Get a line of text from the ini file. Because we have
                eaten all white space, there will be no white space at the
                beginning of the line. */ 
		for(i=0;i<STRING_SIZE-1 && (c=getc(ini_file_pointer))!='\n' &&
                        c!='\r' && c!='\f' && c!='\v' && c!=EOF;i++)line[i]=c;
		line[i]='\0';
		while(c!='\n' && c!='\r' && c!='\f' && c!='\v' && c!=EOF)
			c=getc(ini_file_pointer);

                /* Since a ';' introduces a comment, the part of the 
                line we are interested in, ends at the first ';' character. */ 
                for(i=0;line[i]!=';' && line[i]!='\0';i++);
                line[i]='\0';

                /* If there is only comment at the line, continue with the
                next line. */
                if(line[0]=='\0')continue;

                /* If the first character at the line is a '[', the
                line is a section header. */
                if(line[0]=='[')
                {

                        /* Copy the text between '[' and ']' excluding
                        the '[' and ']', to section_name. */
                        for(i=1;line[i]!='\0' && line[i]!=']';i++)
                                section_name[i-1]=line[i];
                        section_name[i-1]='\0';
                        
                        /* Continue with the next line. */
                        continue;
                }
                
                /* We are only interested in the line, if we are in the servers
                section. */
                if(case_independant_strcmp(section_name,
                        section)!=0)continue;

                /* Read the item name. */
                for(i=0;line[i]!='=' && line[i]!='\0'; i++)
                        item_name[i]=line[i];
                item_name[i]='\0';

                /* Discard all trailing white space at the end of the
                item name. */
                for(i--;i>=0 && (item_name[i]==' ' ||
                        item_name[i]=='\t');i--)
			item_name[i]='\0';

		/* We are only interested in the item, if it is the requested
		item. */
                if(item!=NULL &&
			case_independant_strcmp(item_name,item)!=0)continue;

		/* If a list of items in a section is requested, add the 
		item found. */
		if(item==NULL)
		{
			if(item_name[0]!='\0' && current_size_of_result+
				strlen(item_name)+2<=size_of_result)
			{
				strcpy(result+current_size_of_result,item_name);
				current_size_of_result+=strlen(item_name)+1;
				result[current_size_of_result]='\0';
			}
			continue;
		}

                /* Find the text following the '='. */
                for(i=0;line[i]!='=' && line[i]!='\0'; i++);
		i++;

		/* Discard all leading white space. */
		while(line[i]==' ' || line[i]=='\t')i++;

		/* Copy the found parameters for item. */
		for(j=0;j<size_of_result-1 && 
			line[i]!='\0';i++,j++)
			result[j]=line[i];
		if(size_of_result>0)result[j]='\0';

                /* Discard all trailing white space at the end of the
                item name. */
                for(i--;i>=0 && (result[i]==' ' ||
                        result[i]=='\t');i--)
			result[i]='\0';

		/* Ready. */
                return strlen(result);

        }
}



/* Compare two strings, ignoring the case. If both strings are equal,
the return value is zero, otherwise non-zero. */
int case_independant_strcmp(string1,string2)
char string1[],string2[];
{
        int i;
        char char1,char2;

        for(i=0;;i++)
        {
               if('A'<=string1[i] && string1[i]<='Z')char1=string1[i]-'A'+'a';
               else char1=string1[i]; 
               if('A'<=string2[i] && string2[i]<='Z')char2=string2[i]-'A'+'a';
               else char2=string2[i]; 
               if(char1!=char2)return char1-char2;
               if(char1=='\0')return 0;
        }
}

